"""Media Asset Module

Defines the MediaAsset class for handling media file paths within assets.
"""

import os
from rapidata.rapidata_client.assets.base_asset import BaseAsset

class MediaAsset(BaseAsset):
    """MediaAsset Class

    Represents a media asset by storing the file path.

    Args:
        path (str): The file system path to the media asset.

    Raises:
        FileNotFoundError: If the provided file path does not exist.
    """

    def __init__(self, path: str):
        """
        Initialize a MediaAsset instance.

        Args:
            path (str): The file system path to the media asset.

        Raises:
            FileNotFoundError: If the provided file path does not exist.
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {path}, please provide a valid local file path.")
        self.path = path



from .rapidata_client import (
    RapidataClient,
    ClassifyWorkflow,
    TranscriptionWorkflow,
    CompareWorkflow,
    FreeTextWorkflow,
    DemographicSelection,
    LabelingSelection,
    ValidationSelection,
    ConditionalValidationSelection,
    CappedSelection,
    NaiveReferee,
    ClassifyEarlyStoppingReferee,
    PrivateTextMetadata,
    PublicTextMetadata,
    PromptMetadata,
    TranscriptionMetadata,
    FeatureFlags,
    CountryCodes,
    MediaAsset,
    TextAsset,
    MultiAsset,
    CountryFilter,
    LanguageFilter,
    UserScoreFilter,
    CampaignFilter,
)



from itertools import zip_longest
import os

from rapidata.api_client.models.datapoint_metadata_model import DatapointMetadataModel
from rapidata.api_client.models.datapoint_metadata_model_metadata_inner import (
    DatapointMetadataModelMetadataInner,
)
from rapidata.api_client.models.upload_text_sources_to_dataset_model import (
    UploadTextSourcesToDatasetModel,
)
from rapidata.rapidata_client.metadata.base_metadata import Metadata
from rapidata.rapidata_client.assets import TextAsset, MediaAsset, MultiAsset
from rapidata.service import LocalFileService
from rapidata.service.openapi_service import OpenAPIService
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm


class RapidataDataset:

    def __init__(self, dataset_id: str, openapi_service: OpenAPIService):
        self.dataset_id = dataset_id
        self.openapi_service = openapi_service
        self.local_file_service = LocalFileService()

    def add_texts(
        self,
        text_assets: list[TextAsset] | list[MultiAsset],
        max_workers: int = 10,
    ):
        for text_asset in text_assets:
            if isinstance(text_asset, MultiAsset):
                assert all(
                    isinstance(asset, TextAsset) for asset in text_asset.assets
                ), "All assets in a MultiAsset must be of type TextAsset."

        def upload_text_datapoint(text_asset: TextAsset | MultiAsset) -> None:
            if isinstance(text_asset, TextAsset):
                texts = [text_asset.text]
            elif isinstance(text_asset, MultiAsset):
                texts = [asset.text for asset in text_asset.assets if isinstance(asset, TextAsset)]
            else:
                raise ValueError(f"Unsupported asset type: {type(text_asset)}")

            model = UploadTextSourcesToDatasetModel(
                datasetId=self.dataset_id,
                textSources=texts
            )

            self.openapi_service.dataset_api.dataset_creat_text_datapoint_post(model)

        total_uploads = len(text_assets)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [
                executor.submit(upload_text_datapoint, text_asset)
                for text_asset in text_assets
            ]

            with tqdm(total=total_uploads, desc="Uploading text datapoints") as pbar:
                for future in as_completed(futures):
                    future.result()  # This will raise any exceptions that occurred during execution
                    pbar.update(1)

    def add_media_from_paths(
        self,
        media_paths: list[MediaAsset] |  list[MultiAsset], # where multiasset is a list of media assets
        metadata: list[Metadata] | None = None,
        max_workers: int = 10,
    ):
        if metadata is not None and len(metadata) != len(media_paths):
            raise ValueError(
                "metadata must be None or have the same length as media_paths"
            )

        for media_path in media_paths:
            if isinstance(media_path, MultiAsset):
                assert all(
                    isinstance(asset, MediaAsset) for asset in media_path.assets
                ), "All assets in a MultiAsset must be of type MediaAsset."

        def upload_datapoint(media_asset: MediaAsset | MultiAsset, meta: Metadata | None) -> None:
            if isinstance(media_asset, MediaAsset):
                paths = [media_asset.path]
            elif isinstance(media_asset, MultiAsset):
                paths = [asset.path for asset in media_asset.assets if isinstance(asset, MediaAsset)]
            else:
                raise ValueError(f"Unsupported asset type: {type(media_asset)}")

            assert all(
                os.path.exists(media_path) for media_path in paths
            ), "All media paths must exist on the local filesystem."

            meta_model = meta.to_model() if meta else None
            model = DatapointMetadataModel(
                datasetId=self.dataset_id,
                metadata=(
                    [DatapointMetadataModelMetadataInner(meta_model)]
                    if meta_model
                    else []
                ),
            )

            self.openapi_service.dataset_api.dataset_create_datapoint_post(
                model=model,
                files=paths # type: ignore
            )

        total_uploads = len(media_paths)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [
                executor.submit(upload_datapoint, media_asset, meta)
                for media_asset, meta in zip_longest(media_paths, metadata or [])
            ]

            with tqdm(total=total_uploads, desc="Uploading datapoints") as pbar:
                for future in as_completed(futures):
                    future.result()  # This will raise any exceptions that occurred during execution
                    pbar.update(1)



from typing import Any
from rapidata.rapidata_client.filter.base_filter import Filter
from rapidata.api_client.models.language_user_filter_model import (
    LanguageUserFilterModel,
)


class LanguageFilter(Filter):

    def __init__(self, language_codes: list[str]):
        # check that all characters in the language codes are lowercase
        if not all([code.islower() for code in language_codes]):
            raise ValueError("Language codes must be lowercase")

        self.languages = language_codes

    def to_model(self):
        return LanguageUserFilterModel(_t="LanguageFilter", languages=self.languages)



from typing import Any
from rapidata.rapidata_client.selection.base_selection import Selection
from rapidata.api_client.models.labeling_selection import (
    LabelingSelection as LabelingSelectionModel,
)


class LabelingSelection(Selection):

    def __init__(self, amount: int):
        self.amount = amount

    def to_model(self) -> Any:
        return LabelingSelectionModel(_t="LabelingSelection", amount=self.amount)



from typing import Any
from rapidata.rapidata_client.filter.base_filter import Filter
from rapidata.api_client.models.country_user_filter_model import CountryUserFilterModel


class CountryFilter(Filter):

    def __init__(self, country_codes: list[str]):
        # check that all characters in the country codes are uppercase
        if not all([code.isupper() for code in country_codes]):
            raise ValueError("Country codes must be uppercase")

        self.country_codes = country_codes

    def to_model(self):
        return CountryUserFilterModel(_t="CountryFilter", countries=self.country_codes)



"""Assets Package

This package provides classes for different types of assets, including MediaAsset, TextAsset, and MultiAsset.
"""

from .media_asset import MediaAsset
from .text_asset import TextAsset
from .multi_asset import MultiAsset



from typing import Any
from rapidata.rapidata_client.filter.base_filter import Filter
from rapidata.api_client.models.gender_user_filter_model import GenderUserFilterModel
from rapidata.api_client.models.gender import Gender


class GenderFilter(Filter):

    def __init__(self, genders: list[Gender]):
        self.genders = genders

    def to_model(self):
        return GenderUserFilterModel(
            _t="GenderFilter",
            genders=self.genders,
        )



from rapidata.api_client.models.naive_referee_model import NaiveRefereeModel
from rapidata.rapidata_client.referee.base_referee import Referee


class NaiveReferee(Referee):
    """A simple referee that completes a task after a fixed number of guesses.

    This referee implements a straightforward approach to task completion,
    where the task is considered finished after a predetermined number of
    guesses have been made, regardless of the content or quality of those guesses.

    Attributes:
        required_guesses (int): The number of guesses required to complete the task.
    """

    def __init__(self, required_guesses: int = 10):
        """Initialize the NaiveReferee.

        Args:
            required_guesses (int, optional): The number of guesses required
                to complete the task. Defaults to 10.
        """
        super().__init__()
        self.required_guesses = required_guesses

    def to_dict(self):
        return {
            "_t": "NaiveRefereeConfig",
            "guessesRequired": self.required_guesses,
        }

    def to_model(self):
        return NaiveRefereeModel(_t="NaiveReferee", totalVotes=self.required_guesses)



from rapidata.api_client.models.feature_flag_model import FeatureFlagModel
from enum import StrEnum

class TranslationBehaviour(StrEnum):
    BOTH = "both"
    ONLY_ORIGINAL = "only original"
    ONLY_TRANSLATED = "only translated"


class FeatureFlags:
    """A class to manage feature flags.

    This class provides methods to set and manage various feature flags
    used in the application.

    Attributes:
        _flags (dict[str, str]): A dictionary to store feature flags.
    """

    def __init__(self):
        """Initialize the FeatureFlags object with an empty flags dictionary."""
        self._flags: dict[str, str] = {}

    def to_list(self) -> list[FeatureFlagModel]:
        """Convert the feature flags to a list of FeatureFlagModel objects.

        Returns:
            list[FeatureFlagModel]: A list of FeatureFlagModel objects
                representing the current feature flags.
        """
        return [
            FeatureFlagModel(key=name, value=value)
            for name, value in self._flags.items()
        ]

    def alert_on_fast_response(self, milliseconds: int):
        """Gives an alert as a pop up on the UI when the response time is less than the threshold.

        Args:
            milliseconds (int): if the user responds in less than this time, an alert will be shown.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        if milliseconds < 10:
            print(
                f"Warning: Are you sure you want to set the threshold so low ({milliseconds} milliseconds)?"
            )

        if milliseconds > 30000:
            print(
                f"Warning: Are you sure you want to set the threshold so high ({milliseconds/1000} seconds)?"
            )

        self._flags["alert_on_fast_response"] = str(milliseconds)
        return self


    def translation_behaviour(self, behaviour: TranslationBehaviour = TranslationBehaviour.BOTH):
        """Defines what's the behaviour of the translation in the UI.

        The behaviour can be set to:
            - TranslationBehaviour.BOTH: Show both the original and the translated text.
            - TranslationBehaviour.ONLY_ORIGINAL: Show only the original text.
            - TranslationBehaviour.ONLY_TRANSLATED: Show only the translated text.

        Args:
            behaviour (TranslationBehaviour): The translation behaviour. Defaults to TranslationBehaviour.BOTH.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        self._flags["translation_behaviour"] = behaviour
        return self

    def free_text_minimum_characters(self, value: int):
        """Set the minimum number of characters a user has to type before the task can be successfully submitted.

        Args:
            value (int): The minimum number of characters for free text.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        self._flags["free_text_minimum_characters"] = str(value)
        return self

    def no_shuffle(self, value: bool = True):
        """Only for classify tasks. If true, the order of the categories will be the same as the one in the workflow.

        Args:
            value (bool, optional): Whether to disable shuffling. Defaults to True.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        self._flags["no_shuffle"] = str(value)
        return self

    def compare_with_prompt_design(self, value: bool = True):
        """A special design to compare two texts/images based on a criteria and a given prompt.

        Args:
            value (bool, optional): Whether to enable compare with prompt design. Defaults to True.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        self._flags["claire"] = str(value)
        return self

    def key_value(self, key: str, value: str):
        """Set a custom feature flag with the given key and value. Use this to enable features that do not have a dedicated method (yet).

        Args:
            key (str): The key for the custom feature flag.
            value (str): The value for the custom feature flag.

        Returns:
            FeatureFlags: The current FeatureFlags instance for method chaining.
        """
        self._flags[key] = value
        return self


from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.metadata import Metadata
from rapidata.rapidata_client.order.rapidata_order_builder import RapidataOrderBuilder
from rapidata.rapidata_client.workflow.compare_workflow import CompareWorkflow
from rapidata.rapidata_client.referee.naive_referee import NaiveReferee
from rapidata.rapidata_client.selection.validation_selection import ValidationSelection
from rapidata.rapidata_client.selection.labeling_selection import LabelingSelection
from rapidata.rapidata_client.selection.base_selection import Selection
from rapidata.rapidata_client.assets import MultiAsset, MediaAsset
from typing import Sequence

class CompareOrderBuilder:
    def __init__(self, name:str, criteria: str, media_paths: list[list[str]], openapi_service: OpenAPIService):
        self._order_builder = RapidataOrderBuilder(name=name, openapi_service=openapi_service)
        self._name = name
        self._criteria = criteria
        self._media_paths = media_paths
        self._responses_required = 10
        self._metadata = None
        self._validation_set_id = None

    def responses(self, responses_required: int) -> 'CompareOrderBuilder':
        """Set the number of resoonses required per matchup/pairing for the comparison order."""
        self._responses_required = responses_required
        return self
    
    def metadata(self, metadata: Sequence[Metadata]) -> 'CompareOrderBuilder':
        """Set the metadata for the comparison order. Has to be the same shape as the media paths."""
        self._metadata = metadata
        return self
    
    def validation_set_id(self, validation_set_id: str) -> 'CompareOrderBuilder':
        """Set the validation set ID for the comparison order."""
        self._validation_set_id = validation_set_id
        return self
    
    def create(self, submit: bool = True, max_upload_workers: int = 10):
        selection: list[Selection] = ([ValidationSelection(amount=1, validation_set_id=self._validation_set_id), LabelingSelection(amount=2)] 
                     if self._validation_set_id 
                     else [LabelingSelection(amount=3)])
        
        media_paths = [MultiAsset([MediaAsset(path=path) for path in paths]) for paths in self._media_paths]
        order = (self._order_builder
            .workflow(
                CompareWorkflow(
                    criteria=self._criteria
                )
            )
            .referee(NaiveReferee(required_guesses=self._responses_required))
            .media(media_paths, metadata=self._metadata) # type: ignore
            .selections(selection)
            .create(submit=submit, max_workers=max_upload_workers))
        
        return order

class CompareMediaBuilder:
    def __init__(self, name: str, criteria: str, openapi_service: OpenAPIService):
        self._openapi_service = openapi_service
        self._name = name
        self._criteria = criteria
        self._media_paths = None

    def media(self, media_paths: list[list[str]]) -> CompareOrderBuilder:
        """Set the media assets for the comparison order by providing the local paths to the files."""
        self._media_paths = media_paths
        return self._build()
    
    def _build(self) -> CompareOrderBuilder:
        if self._media_paths is None:
            raise ValueError("Media paths are required")
        assert all([len(path) == 2 for path in self._media_paths]), "The media paths must come in pairs for comparison tasks."
        return CompareOrderBuilder(self._name, self._criteria, self._media_paths, self._openapi_service)

class CompareCriteriaBuilder:
    def __init__(self, name: str, openapi_service: OpenAPIService):
        self._openapi_service = openapi_service
        self._name = name
        self._criteria = None

    def criteria(self, criteria: str) -> CompareMediaBuilder:
        """Set the criteria how the images should be compared."""
        self._criteria = criteria
        return self._build()
    
    def _build(self) -> CompareMediaBuilder:
        if self._criteria is None:
            raise ValueError("Criteria are required")
        return CompareMediaBuilder(self._name, self._criteria, self._openapi_service)




from typing import Any
from rapidata.rapidata_client.filter.base_filter import Filter
from rapidata.api_client.models.campaign_user_filter_model import (
    CampaignUserFilterModel,
)


class CampaignFilter(Filter):

    def __init__(self, campaign_ids: list[str]):
        self.campaign_ids = campaign_ids

    def to_model(self):
        return CampaignUserFilterModel(
            _t="CampaignFilter",
            campaignIds=self.campaign_ids,
        )



from rapidata.rapidata_client.selection.base_selection import Selection
from rapidata.api_client.models.conditional_validation_rapid_selection_config import (
    ValidationChance,
)
from rapidata.api_client.models.conditional_validation_selection import (
    ConditionalValidationSelection as ConditionalValidationSelectionModel,
)


class ConditionalValidationSelection(Selection):
    """Demographic selection class."""

    def __init__(
        self,
        validation_set_id: str,
        thresholds: list[float],
        chances: list[float],
        rapid_counts: list[int],
    ):
        self.validation_set_id = validation_set_id
        self.thresholds = thresholds
        self.chances = chances
        self.rapid_counts = rapid_counts

    def to_model(self):
        return ConditionalValidationSelectionModel(
            _t="ConditionalValidationSelection",
            validationSetId=self.validation_set_id,
            validationChances=[
                ValidationChance(
                    userScoreThreshold=threshold, chance=chance, rapidCount=rapid_count
                )
                for threshold, chance, rapid_count in zip(
                    self.thresholds, self.chances, self.rapid_counts
                )
            ],
        )



from io import BytesIO
import PIL.Image as Image

class ImageUtils:

    @staticmethod
    def convert_PIL_image_to_bytes(image: Image.Image):
        """
        Convert a PIL image to bytes with meta data encoded. We can't just use image.tobytes() because this only returns the pixel data.
        """
        buffer = BytesIO()
        image.save(buffer, image.format)
        return buffer.getvalue()